@@ -16,7 +16,7 @@
 
 
 #define PLUGIN_VERSION 0x104
-#define PLUGIN_STR_VERSION "1.4.8"
+#define PLUGIN_STR_VERSION "1.4.10"
 
 #define _my_thread_var loc_thread_var
 
@@ -327,6 +327,10 @@ struct connection_info
   char query_buffer[1024];
   time_t query_time;
   int log_always;
+  char proxy[64];
+  int proxy_length;
+  char proxy_host[64];
+  int proxy_host_length;
 };
 
 #define DEFAULT_FILENAME_LEN 16
@@ -1131,9 +1135,13 @@ static void setup_connection_simple(struct connection_info *ci)
   ci->ip_length= 0;
   ci->query_length= 0;
   ci->header= 0;
+  ci->proxy_length= 0;
 }
 
 
+#define MAX_HOSTNAME 61
+#define USERNAME_LENGTH 384
+
 static void setup_connection_connect(struct connection_info *cn,
     const struct mysql_event_connection *event)
 {
@@ -1150,6 +1158,29 @@ static void setup_connection_connect(struct connection_info *cn,
   get_str_n(cn->ip, &cn->ip_length, sizeof(cn->ip),
             event->ip, event->ip_length);
   cn->header= 0;
+  if (event->proxy_user && event->proxy_user[0])
+  {
+    const char *priv_host= event->proxy_user +
+            sizeof(char[MAX_HOSTNAME+USERNAME_LENGTH+5]);
+    size_t priv_host_length;
+
+    if (mysql_57_started)
+    {
+      priv_host+= sizeof(size_t);
+      priv_host_length= *(size_t *) (priv_host + MAX_HOSTNAME);
+    }
+    else
+      priv_host_length= strlen(priv_host);
+
+
+    get_str_n(cn->proxy, &cn->proxy_length, sizeof(cn->proxy),
+              event->priv_user, event->priv_user_length);
+    get_str_n(cn->proxy_host, &cn->proxy_host_length,
+              sizeof(cn->proxy_host),
+              priv_host, priv_host_length);
+  }
+  else
+    cn->proxy_length= 0;
 }
 
 
@@ -1349,6 +1380,31 @@ static size_t log_header(char *message, size_t message_len,
 }
 
 
+static int log_proxy(const struct connection_info *cn,
+                     const struct mysql_event_connection *event)
+                   
+{
+  time_t ctime;
+  size_t csize;
+  char message[1024];
+
+  (void) time(&ctime);
+  csize= log_header(message, sizeof(message)-1, &ctime,
+                    servhost, servhost_len,
+                    cn->user, cn->user_length,
+                    cn->host, cn->host_length,
+                    cn->ip, cn->ip_length,
+                    event->thread_id, 0, "PROXY_CONNECT");
+  csize+= my_snprintf(message+csize, sizeof(message) - 1 - csize,
+    ",%.*s,`%.*s`@`%.*s`,%d", cn->db_length, cn->db,
+                     cn->proxy_length, cn->proxy,
+                     cn->proxy_host_length, cn->proxy_host,
+                     event->status);
+  message[csize]= '\n';
+  return write_log(message, csize + 1, 1);
+}
+
+
 static int log_connection(const struct connection_info *cn,
                           const struct mysql_event_connection *event,
                           const char *type)
@@ -1431,7 +1487,6 @@ static size_t escape_string_hide_passwords(const char *str, unsigned int len,
   const char *res_start= result;
   const char *res_end= result + result_len - 2;
   size_t d_len;
-  char b_char;
 
   while (len)
   {
@@ -1469,27 +1524,28 @@ static size_t escape_string_hide_passwords(const char *str, unsigned int len,
 
       if (*next_s)
       {
-        memmove(result + d_len, "*****", 5);
+        const char b_char= *next_s++;
+        memset(result + d_len, '*', 5);
         result+= d_len + 5;
-        b_char= *(next_s++);
+
+        while (*next_s)
+        {
+          if (*next_s == b_char)
+          {
+            ++next_s;
+            break;
+          }
+          if (*next_s == '\\')
+          {
+            if (next_s[1])
+              next_s++;
+          }
+          next_s++;
+        }
       }
       else
         result+= d_len;
 
-      while (*next_s)
-      {
-        if (*next_s == b_char)
-        {
-          ++next_s;
-          break;
-        }
-        if (*next_s == '\\')
-        {
-          if (next_s[1])
-            next_s++;
-        }
-        next_s++;
-      }
       len-= (uint)(next_s - str);
       str= next_s;
       continue;
@@ -1497,19 +1553,23 @@ static size_t escape_string_hide_passwords(const char *str, unsigned int len,
 no_password:
     if (result >= res_end)
       break;
-    if ((b_char= escaped_char(*str)))
-    {
-      if (result+1 >= res_end)
-        break;
-      *(result++)= '\\';
-      *(result++)= b_char;
-    }
-    else if (is_space(*str))
-      *(result++)= ' ';
     else
-      *(result++)= *str;
-    str++;
-    len--;
+    {
+      const char b_char= escaped_char(*str);
+      if (b_char)
+      {
+        if (result+1 >= res_end)
+          break;
+        *(result++)= '\\';
+        *(result++)= b_char;
+      }
+      else if (is_space(*str))
+        *(result++)= ' ';
+      else
+        *(result++)= *str;
+      str++;
+      len--;
+    }
   }
   *result= 0;
   return result - res_start;
@@ -2006,9 +2066,13 @@ static void update_connection_info(struct connection_info *cn,
     {
       case MYSQL_AUDIT_CONNECTION_CONNECT:
         setup_connection_connect(cn, event);
+        if (event->status == 0 && event->proxy_user && event->proxy_user[0])
+          log_proxy(cn, event);
         break;
       case MYSQL_AUDIT_CONNECTION_CHANGE_USER:
         *after_action= AA_CHANGE_USER;
+        if (event->proxy_user && event->proxy_user[0])
+          log_proxy(cn, event);
         break;
       default:;
     }
