@@ -2,7 +2,7 @@
 
 Copyright (c) 1996, 2016, Oracle and/or its affiliates. All Rights Reserved.
 Copyright (c) 2008, Google Inc.
-Copyright (c) 2017, 2022, MariaDB Corporation.
+Copyright (c) 2017, 2020, MariaDB Corporation.
 
 Portions of this file contain modifications contributed and copyrighted by
 Google, Inc. Those modifications are gratefully acknowledged and are described
@@ -289,18 +289,11 @@ void btr_search_sys_free()
 
 /** Set index->ref_count = 0 on all indexes of a table.
 @param[in,out]	table	table handler */
-static
-void
-btr_search_disable_ref_count(
-	dict_table_t*	table)
+static void btr_search_disable_ref_count(dict_table_t *table)
 {
-	dict_index_t*	index;
-
-	for (index = dict_table_get_first_index(table);
-	     index != NULL;
-	     index = dict_table_get_next_index(index)) {
-		index->search_info->ref_count = 0;
-	}
+  for (dict_index_t *index= dict_table_get_first_index(table); index;
+       index= dict_table_get_next_index(index))
+    index->search_info->ref_count= 0;
 }
 
 /** Lazily free detached metadata when removing the last reference. */
@@ -308,8 +301,6 @@ ATTRIBUTE_COLD static void btr_search_lazy_free(dict_index_t *index)
 {
   ut_ad(index->freed());
   dict_table_t *table= index->table;
-  mysql_mutex_lock(&table->autoinc_mutex);
-
   /* Perform the skipped steps of dict_index_remove_from_cache_low(). */
   UT_LIST_REMOVE(table->freed_indexes, index);
   rw_lock_free(&index->lock);
@@ -318,14 +309,9 @@ ATTRIBUTE_COLD static void btr_search_lazy_free(dict_index_t *index)
   if (!UT_LIST_GET_LEN(table->freed_indexes) &&
       !UT_LIST_GET_LEN(table->indexes))
   {
-    ut_ad(!table->id);
-    mysql_mutex_unlock(&table->autoinc_mutex);
-    mysql_mutex_destroy(&table->autoinc_mutex);
+    ut_ad(table->id == 0);
     dict_mem_table_free(table);
-    return;
   }
-
-  mysql_mutex_unlock(&table->autoinc_mutex);
 }
 
 /** Clear the adaptive hash index on all pages in the buffer pool. */
@@ -390,12 +376,12 @@ void btr_search_disable()
 {
 	dict_table_t*	table;
 
-	mutex_enter(&dict_sys->mutex);
+	mutex_enter(&dict_sys.mutex);
 
 	btr_search_x_lock_all();
 
 	if (!btr_search_enabled) {
-		mutex_exit(&dict_sys->mutex);
+		mutex_exit(&dict_sys.mutex);
 		btr_search_x_unlock_all();
 		return;
 	}
@@ -404,19 +390,19 @@ void btr_search_disable()
 
 	/* Clear the index->search_info->ref_count of every index in
 	the data dictionary cache. */
-	for (table = UT_LIST_GET_FIRST(dict_sys->table_LRU); table;
+	for (table = UT_LIST_GET_FIRST(dict_sys.table_LRU); table;
 	     table = UT_LIST_GET_NEXT(table_LRU, table)) {
 
 		btr_search_disable_ref_count(table);
 	}
 
-	for (table = UT_LIST_GET_FIRST(dict_sys->table_non_LRU); table;
+	for (table = UT_LIST_GET_FIRST(dict_sys.table_non_LRU); table;
 	     table = UT_LIST_GET_NEXT(table_LRU, table)) {
 
 		btr_search_disable_ref_count(table);
 	}
 
-	mutex_exit(&dict_sys->mutex);
+	mutex_exit(&dict_sys.mutex);
 
 	/* Set all block->index = NULL. */
 	buf_pool_clear_hash_index();
@@ -680,12 +666,6 @@ btr_search_update_hash_ref(
 		return;
 	}
 
-	if (cursor->index != index) {
-		ut_ad(cursor->index->id == index->id);
-		btr_search_drop_page_hash_index(block);
-		return;
-	}
-
 	ut_ad(block->page.id.space() == index->table->space_id);
 	ut_ad(index == cursor->index);
 	ut_ad(!dict_index_is_ibuf(index));
@@ -710,8 +690,7 @@ btr_search_update_hash_ref(
 
 		ulint fold = rec_fold(
 			rec,
-			rec_get_offsets(rec, index, offsets_,
-					index->n_core_fields,
+			rec_get_offsets(rec, index, offsets_, true,
 					ULINT_UNDEFINED, &heap),
 			block->curr_n_fields,
 			block->curr_n_bytes, index->id);
@@ -770,8 +749,7 @@ btr_search_check_guess(
 
 	match = 0;
 
-	offsets = rec_get_offsets(rec, cursor->index, offsets,
-				  cursor->index->n_core_fields,
+	offsets = rec_get_offsets(rec, cursor->index, offsets, true,
 				  n_unique, &heap);
 	cmp = cmp_dtuple_rec_with_match(tuple, rec, offsets, &match);
 
@@ -822,8 +800,7 @@ btr_search_check_guess(
 		}
 
 		offsets = rec_get_offsets(prev_rec, cursor->index, offsets,
-					  cursor->index->n_core_fields,
-					  n_unique, &heap);
+					  true, n_unique, &heap);
 		cmp = cmp_dtuple_rec_with_match(
 			tuple, prev_rec, offsets, &match);
 		if (mode == PAGE_CUR_GE) {
@@ -846,8 +823,7 @@ btr_search_check_guess(
 		}
 
 		offsets = rec_get_offsets(next_rec, cursor->index, offsets,
-					  cursor->index->n_core_fields,
-					  n_unique, &heap);
+					  true, n_unique, &heap);
 		cmp = cmp_dtuple_rec_with_match(
 			tuple, next_rec, offsets, &match);
 		if (mode == PAGE_CUR_LE) {
@@ -895,6 +871,8 @@ both have sensible values.
 				we assume the caller uses his search latch
 				to protect the record!
 @param[out]	cursor		tree cursor
+@param[in]	ahi_latch	the adaptive hash index latch being held,
+				or NULL
 @param[in]	mtr		mini transaction
 @return whether the search succeeded */
 bool
@@ -905,6 +883,7 @@ btr_search_guess_on_hash(
 	ulint		mode,
 	ulint		latch_mode,
 	btr_cur_t*	cursor,
+	rw_lock_t*	ahi_latch,
 	mtr_t*		mtr)
 {
 	ulint		fold;
@@ -913,6 +892,8 @@ btr_search_guess_on_hash(
 	btr_cur_t	cursor2;
 	btr_pcur_t	pcur;
 #endif
+	ut_ad(!ahi_latch || rw_lock_own_flagged(
+		      ahi_latch, RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
 
 	if (!btr_search_enabled) {
 		return false;
@@ -920,6 +901,7 @@ btr_search_guess_on_hash(
 
 	ut_ad(index && info && tuple && cursor && mtr);
 	ut_ad(!dict_index_is_ibuf(index));
+	ut_ad(!ahi_latch || ahi_latch == btr_get_search_latch(index));
 	ut_ad((latch_mode == BTR_SEARCH_LEAF)
 	      || (latch_mode == BTR_MODIFY_LEAF));
 
@@ -950,21 +932,28 @@ btr_search_guess_on_hash(
 	cursor->fold = fold;
 	cursor->flag = BTR_CUR_HASH;
 
-	rw_lock_t* ahi_latch = btr_get_search_latch(index);
+	rw_lock_t* use_latch = ahi_latch ? NULL : btr_get_search_latch(index);
 	const rec_t* rec;
 
-		rw_lock_s_lock(ahi_latch);
+	if (use_latch) {
+		rw_lock_s_lock(use_latch);
 
-	if (!btr_search_enabled) {
-		goto fail;
+		if (!btr_search_enabled) {
+			goto fail;
+		}
+	} else {
+		ut_ad(btr_search_enabled);
+		ut_ad(rw_lock_own(ahi_latch, RW_LOCK_S));
 	}
 
 	rec = static_cast<const rec_t*>(
 		ha_search_and_get_data(btr_get_search_table(index), fold));
 
 	if (!rec) {
+		if (use_latch) {
 fail:
-		rw_lock_s_unlock(ahi_latch);
+			rw_lock_s_unlock(use_latch);
+		}
 
 		btr_search_failure(info, cursor);
 		return false;
@@ -972,19 +961,25 @@ fail:
 
 	buf_block_t*	block = buf_block_from_ahi(rec);
 
-	if (!buf_page_get_known_nowait(latch_mode, block, BUF_MAKE_YOUNG,
-	      __FILE__, __LINE__, mtr)) {
-		goto fail;
-	}
+	if (use_latch) {
+		if (!buf_page_get_known_nowait(
+			latch_mode, block, BUF_MAKE_YOUNG,
+			__FILE__, __LINE__, mtr)) {
+			goto fail;
+		}
 
-	const bool fail = index != block->index
-		&& index_id == block->index->id;
-	ut_a(!fail || block->index->freed());
-	ut_ad(fail || !block->page.file_page_was_freed);
-	rw_lock_s_unlock(ahi_latch);
+		const bool fail = index != block->index
+			&& index_id == block->index->id;
+		ut_a(!fail || block->index->freed());
+		rw_lock_s_unlock(use_latch);
 
-	buf_block_dbg_add_level(block, SYNC_TREE_NODE_FROM_HASH);
-	if (UNIV_UNLIKELY(fail)) {
+		buf_block_dbg_add_level(block, SYNC_TREE_NODE_FROM_HASH);
+		if (UNIV_UNLIKELY(fail)) {
+			goto fail_and_release_page;
+		}
+	} else if (UNIV_UNLIKELY(index != block->index
+				 && index_id == block->index->id)) {
+		ut_a(block->index->freed());
 		goto fail_and_release_page;
 	}
 
@@ -993,7 +988,9 @@ fail:
 		ut_ad(buf_block_get_state(block) == BUF_BLOCK_REMOVE_HASH);
 
 fail_and_release_page:
-		btr_leaf_page_release(block, latch_mode, mtr);
+		if (!ahi_latch) {
+			btr_leaf_page_release(block, latch_mode, mtr);
+		}
 
 		btr_search_failure(info, cursor);
 		return false;
@@ -1011,7 +1008,7 @@ fail_and_release_page:
 	record to determine if our guess for the cursor position is
 	right. */
 	if (index_id != btr_page_get_index_id(block->frame)
-	    || !btr_search_check_guess(cursor, 0, tuple, mode)) {
+	    || !btr_search_check_guess(cursor, !!ahi_latch, tuple, mode)) {
 		goto fail_and_release_page;
 	}
 
@@ -1060,7 +1057,7 @@ fail_and_release_page:
 #ifdef UNIV_SEARCH_PERF_STAT
 	btr_search_n_succ++;
 #endif
-	if (buf_page_peek_if_too_old(&block->page)) {
+	if (!ahi_latch && buf_page_peek_if_too_old(&block->page)) {
 
 		buf_page_make_young(&block->page);
 	}
@@ -1081,11 +1078,8 @@ fail_and_release_page:
 			index page for which we know that
 			block->buf_fix_count == 0 or it is an index page which
 			has already been removed from the buf_pool->page_hash
-			i.e.: it is in state BUF_BLOCK_REMOVE_HASH
-@param[in]	garbage_collect	drop ahi only if the index is marked
-				as freed */
-void btr_search_drop_page_hash_index(buf_block_t* block,
-				     bool garbage_collect)
+			i.e.: it is in state BUF_BLOCK_REMOVE_HASH */
+void btr_search_drop_page_hash_index(buf_block_t* block)
 {
 	ulint			n_fields;
 	ulint			n_bytes;
@@ -1102,6 +1096,9 @@ void btr_search_drop_page_hash_index(buf_block_t* block,
 	rw_lock_t*		latch;
 
 retry:
+	/* This debug check uses a dirty read that could theoretically cause
+	false positives while buf_pool_clear_hash_index() is executing. */
+	assert_block_ahi_valid(block);
 	ut_ad(!btr_search_own_any(RW_LOCK_S));
 	ut_ad(!btr_search_own_any(RW_LOCK_X));
 
@@ -1117,7 +1114,7 @@ retry:
 	ut_ad(page_is_leaf(block->frame));
 
 	/* We must not dereference block->index here, because it could be freed
-	if (index->table->n_ref_count == 0 && !mutex_own(&dict_sys->mutex)).
+	if (index->table->n_ref_count == 0 && !mutex_own(&dict_sys.mutex)).
 	Determine the ahi_slot based on the block contents. */
 
 	const index_id_t	index_id
@@ -1129,33 +1126,14 @@ retry:
 	latch = btr_search_latches[ahi_slot];
 
 	rw_lock_s_lock(latch);
-
-	dict_index_t* index = block->index;
-
-	bool is_freed = index && index->freed();
-	if (is_freed) {
-		rw_lock_s_unlock(latch);
-		rw_lock_x_lock(latch);
-		if (index != block->index) {
-			rw_lock_x_unlock(latch);
-			goto retry;
-		}
-	} else if (garbage_collect) {
-		rw_lock_s_unlock(latch);
-		return;
-	}
-
 	assert_block_ahi_valid(block);
 
-	if (!index || !btr_search_enabled) {
-		if (is_freed) {
-			rw_lock_x_unlock(latch);
-		} else {
-			rw_lock_s_unlock(latch);
-		}
+	if (!block->index || !btr_search_enabled) {
+		rw_lock_s_unlock(latch);
 		return;
 	}
 
+	dict_index_t* index = block->index;
 #ifdef MYSQL_INDEX_DISABLE_AHI
 	ut_ad(!index->disable_ahi);
 #endif
@@ -1171,9 +1149,7 @@ retry:
 	/* NOTE: The AHI fields of block must not be accessed after
 	releasing search latch, as the index page might only be s-latched! */
 
-	if (!is_freed) {
-		rw_lock_s_unlock(latch);
-	}
+	rw_lock_s_unlock(latch);
 
 	ut_a(n_fields > 0 || n_bytes > 0);
 
@@ -1189,7 +1165,7 @@ retry:
 
 	rec = page_get_infimum_rec(page);
 	rec = page_rec_get_next_low(rec, page_is_comp(page));
-	if (rec_is_metadata(rec, index)) {
+	if (rec_is_metadata(rec, *index)) {
 		rec = page_rec_get_next_low(rec, page_is_comp(page));
 	}
 
@@ -1200,7 +1176,7 @@ retry:
 
 	while (!page_rec_is_supremum(rec)) {
 		offsets = rec_get_offsets(
-			rec, index, offsets, index->n_core_fields,
+			rec, index, offsets, true,
 			btr_search_get_n_fields(n_fields, n_bytes),
 			&heap);
 		fold = rec_fold(rec, offsets, n_fields, n_bytes, index_id);
@@ -1224,18 +1200,16 @@ next_rec:
 		mem_heap_free(heap);
 	}
 
-	if (!is_freed) {
-		rw_lock_x_lock(latch);
+	rw_lock_x_lock(latch);
 
-		if (UNIV_UNLIKELY(!block->index)) {
-			/* Someone else has meanwhile dropped the
-			hash index */
-			goto cleanup;
-		}
+	if (UNIV_UNLIKELY(!block->index)) {
+		/* Someone else has meanwhile dropped the hash index */
 
-		ut_a(block->index == index);
+		goto cleanup;
 	}
 
+	ut_a(block->index == index);
+
 	if (block->curr_n_fields != n_fields
 	    || block->curr_n_bytes != n_bytes) {
 
@@ -1293,7 +1267,7 @@ void btr_search_drop_page_hash_when_freed(const page_id_t page_id)
 	are possibly holding, we cannot s-latch the page, but must
 	(recursively) x-latch it, even though we are only reading. */
 
-	block = buf_page_get_gen(page_id, univ_page_size, RW_X_LATCH, NULL,
+	block = buf_page_get_gen(page_id, 0, RW_X_LATCH, NULL,
 				 BUF_PEEK_IF_IN_POOL, __FILE__, __LINE__,
 				 &mtr, &err);
 
@@ -1311,7 +1285,7 @@ void btr_search_drop_page_hash_when_freed(const page_id_t page_id)
 			be open, or we should be in the process of
 			dropping the table (preventing eviction). */
 			ut_ad(index->table->get_ref_count() > 0
-			      || mutex_own(&dict_sys->mutex));
+			      || mutex_own(&dict_sys.mutex));
 			btr_search_drop_page_hash_index(block);
 		}
 	}
@@ -1409,7 +1383,7 @@ btr_search_build_page_hash_index(
 
 	rec = page_rec_get_next_const(page_get_infimum_rec(page));
 
-	if (rec_is_metadata(rec, index)) {
+	if (rec_is_metadata(rec, *index)) {
 		rec = page_rec_get_next_const(rec);
 		if (!--n_recs) return;
 	}
@@ -1426,7 +1400,7 @@ btr_search_build_page_hash_index(
 	ut_a(index->id == btr_page_get_index_id(page));
 
 	offsets = rec_get_offsets(
-		rec, index, offsets, index->n_core_fields,
+		rec, index, offsets, true,
 		btr_search_get_n_fields(n_fields, n_bytes),
 		&heap);
 	ut_ad(page_rec_is_supremum(rec)
@@ -1457,7 +1431,7 @@ btr_search_build_page_hash_index(
 		}
 
 		offsets = rec_get_offsets(
-			next_rec, index, offsets, index->n_core_fields,
+			next_rec, index, offsets, true,
 			btr_search_get_n_fields(n_fields, n_bytes), &heap);
 		next_fold = rec_fold(next_rec, offsets, n_fields,
 				     n_bytes, index->id);
@@ -1609,7 +1583,6 @@ btr_search_move_or_delete_hash_entries(
 	rw_lock_t* ahi_latch = index ? btr_get_search_latch(index) : NULL;
 
 	if (new_block->index) {
-drop_exit:
 		btr_search_drop_page_hash_index(block);
 		return;
 	}
@@ -1621,12 +1594,6 @@ drop_exit:
 	rw_lock_s_lock(ahi_latch);
 
 	if (block->index) {
-
-		if (block->index != index) {
-			rw_lock_s_unlock(ahi_latch);
-			goto drop_exit;
-		}
-
 		ulint	n_fields = block->curr_n_fields;
 		ulint	n_bytes = block->curr_n_bytes;
 		ibool	left_side = block->curr_left_side;
@@ -1647,6 +1614,7 @@ drop_exit:
 		ut_ad(left_side == block->curr_left_side);
 		return;
 	}
+
 	rw_lock_s_unlock(ahi_latch);
 }
 
@@ -1684,12 +1652,6 @@ void btr_search_update_hash_on_delete(btr_cur_t* cursor)
 		return;
 	}
 
-	if (index != cursor->index) {
-		ut_ad(index->id == cursor->index->id);
-		btr_search_drop_page_hash_index(block);
-		return;
-	}
-
 	ut_ad(block->page.id.space() == index->table->space_id);
 	ut_a(index == cursor->index);
 	ut_a(block->curr_n_fields > 0 || block->curr_n_bytes > 0);
@@ -1697,8 +1659,7 @@ void btr_search_update_hash_on_delete(btr_cur_t* cursor)
 
 	rec = btr_cur_get_rec(cursor);
 
-	fold = rec_fold(rec, rec_get_offsets(rec, index, offsets_,
-					     index->n_core_fields,
+	fold = rec_fold(rec, rec_get_offsets(rec, index, offsets_, true,
 					     ULINT_UNDEFINED, &heap),
 			block->curr_n_fields, block->curr_n_bytes, index->id);
 	if (UNIV_LIKELY_NULL(heap)) {
@@ -1764,12 +1725,6 @@ btr_search_update_hash_node_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 		return;
 	}
 
-	if (cursor->index != index) {
-		ut_ad(cursor->index->id == index->id);
-		btr_search_drop_page_hash_index(block);
-		return;
-	}
-
 	ut_a(cursor->index == index);
 	ut_ad(!dict_index_is_ibuf(index));
 	rw_lock_x_lock(ahi_latch);
@@ -1859,12 +1814,6 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 #ifdef MYSQL_INDEX_DISABLE_AHI
 	ut_a(!index->disable_ahi);
 #endif
-	if (index != cursor->index) {
-		ut_ad(index->id == cursor->index->id);
-		btr_search_drop_page_hash_index(block);
-		return;
-	}
-
 	ut_a(index == cursor->index);
 	ut_ad(!dict_index_is_ibuf(index));
 
@@ -1875,14 +1824,13 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 	ins_rec = page_rec_get_next_const(rec);
 	next_rec = page_rec_get_next_const(ins_rec);
 
-	offsets = rec_get_offsets(ins_rec, index, offsets,
-				  index->n_core_fields,
+	offsets = rec_get_offsets(ins_rec, index, offsets, true,
 				  ULINT_UNDEFINED, &heap);
 	ins_fold = rec_fold(ins_rec, offsets, n_fields, n_bytes, index->id);
 
 	if (!page_rec_is_supremum(next_rec)) {
 		offsets = rec_get_offsets(
-			next_rec, index, offsets, index->n_core_fields,
+			next_rec, index, offsets, true,
 			btr_search_get_n_fields(n_fields, n_bytes), &heap);
 		next_fold = rec_fold(next_rec, offsets, n_fields,
 				     n_bytes, index->id);
@@ -1892,9 +1840,9 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 	hash_table_t* table = NULL;
 	bool locked = false;
 
-	if (!page_rec_is_infimum(rec) && !rec_is_metadata(rec, index)) {
+	if (!page_rec_is_infimum(rec) && !rec_is_metadata(rec, *index)) {
 		offsets = rec_get_offsets(
-			rec, index, offsets, index->n_core_fields,
+			rec, index, offsets, true,
 			btr_search_get_n_fields(n_fields, n_bytes), &heap);
 		fold = rec_fold(rec, offsets, n_fields, n_bytes, index->id);
 	} else {
@@ -2100,8 +2048,7 @@ btr_search_hash_table_validate(ulint hash_table_id)
 			page_index_id = btr_page_get_index_id(block->frame);
 
 			offsets = rec_get_offsets(
-				node->data, block->index, offsets,
-				block->index->n_core_fields,
+				node->data, block->index, offsets, true,
 				btr_search_get_n_fields(block->curr_n_fields,
 							block->curr_n_bytes),
 				&heap);
@@ -2207,22 +2154,5 @@ btr_search_validate()
 	return(true);
 }
 
-#ifdef UNIV_DEBUG
-bool btr_search_check_marked_free_index(const buf_block_t *block)
-{
-  const index_id_t index_id= btr_page_get_index_id(block->frame);
-
-  rw_lock_t *ahi_latch= btr_get_search_latch(
-    index_id, block->page.id.space());
-
-  rw_lock_s_lock(ahi_latch);
-
-  bool is_freed= block->index && block->index->freed();
-
-  rw_lock_s_unlock(ahi_latch);
-
-  return is_freed;
-}
-#endif /* UNIV_DEBUG */
 #endif /* defined UNIV_AHI_DEBUG || defined UNIV_DEBUG */
 #endif /* BTR_CUR_HASH_ADAPT */
