@@ -28,9 +28,7 @@ Created 11/26/1995 Heikki Tuuri
 #ifndef mtr0mtr_h
 #define mtr0mtr_h
 
-#include "log0types.h"
-#include "mtr0types.h"
-#include "buf0types.h"
+#include "fil0fil.h"
 #include "dyn0buf.h"
 
 /** Start a mini-transaction. */
@@ -56,23 +54,12 @@ savepoint. */
 @return	old mode */
 #define mtr_set_log_mode(m, d)	(m)->set_log_mode((d))
 
-/** Read 1 - 4 bytes from a file page buffered in the buffer pool.
-@return	value read */
-#define mtr_read_ulint(p, t, m)	(m)->read_ulint((p), (t))
-
 /** Release an object in the memo stack.
 @return true if released */
 #define mtr_memo_release(m, o, t)					\
 				(m)->memo_release((o), (t))
 
 #ifdef UNIV_DEBUG
-
-/** Check if memo contains the given item. */
-#define mtr_is_block_fix(m, o, t, table) mtr_memo_contains(m, o, t)
-
-/** Check if memo contains the given page. */
-#define mtr_is_page_fix(m, p, t, table) mtr_memo_contains_page(m, p, t)
-
 /** Check if memo contains the given item.
 @return	TRUE if contains */
 #define mtr_memo_contains(m, o, t)					\
@@ -94,17 +81,12 @@ savepoint. */
 /** Push an object to an mtr memo stack. */
 #define mtr_memo_push(m, o, t)	(m)->memo_push(o, t)
 
-/** Lock an rw-lock in s-mode. */
-#define mtr_s_lock(l, m)	(m)->s_lock((l), __FILE__, __LINE__)
-
-/** Lock an rw-lock in x-mode. */
-#define mtr_x_lock(l, m)	(m)->x_lock((l), __FILE__, __LINE__)
-
-/** Lock a tablespace in x-mode. */
+#define mtr_s_lock_space(s, m)	(m)->s_lock_space((s), __FILE__, __LINE__)
 #define mtr_x_lock_space(s, m)	(m)->x_lock_space((s), __FILE__, __LINE__)
 
-/** Lock an rw-lock in sx-mode. */
-#define mtr_sx_lock(l, m)	(m)->sx_lock((l), __FILE__, __LINE__)
+#define mtr_s_lock_index(i, m)	(m)->s_lock(&(i)->lock, __FILE__, __LINE__)
+#define mtr_x_lock_index(i, m)	(m)->x_lock(&(i)->lock, __FILE__, __LINE__)
+#define mtr_sx_lock_index(i, m)	(m)->sx_lock(&(i)->lock, __FILE__, __LINE__)
 
 #define mtr_memo_contains_flagged(m, p, l)				\
 				(m)->memo_contains_flagged((p), (l))
@@ -126,9 +108,6 @@ savepoint. */
 @return true if the mtr is dirtying a clean page. */
 #define mtr_block_dirtied(b)	mtr_t::is_block_dirtied((b))
 
-/** Forward declaration of a tablespace object */
-struct fil_space_t;
-
 /** Append records to the system-wide redo log buffer.
 @param[in]	log	redo log records */
 void
@@ -148,13 +127,6 @@ struct mtr_memo_slot_t {
 struct mtr_t {
 	mtr_t() : m_state(MTR_STATE_INIT) {}
 
-	/** Release the free extents that was reserved using
-	fsp_reserve_free_extents().  This is equivalent to calling
-	fil_space_release_free_extents().  This is intended for use
-	with index pages.
-	@param[in]	n_reserved	number of reserved extents */
-	void release_free_extents(ulint n_reserved);
-
 	/** Start a mini-transaction. */
 	void start();
 
@@ -206,17 +178,6 @@ struct mtr_t {
 	@return	old mode */
 	inline mtr_log_t set_log_mode(mtr_log_t mode);
 
-	/** Note that the mini-transaction is modifying the system tablespace
-	(for example, for the change buffer or for undo logs)
-	@return the system tablespace */
-	fil_space_t* set_sys_modified()
-	{
-		if (!m_sys_space) {
-			lookup_sys_space();
-		}
-		return m_sys_space;
-	}
-
 	/** Copy the tablespaces associated with the mini-transaction
 	(needed for generating MLOG_FILE_NAME records)
 	@param[in]	mtr	mini-transaction that may modify
@@ -225,27 +186,26 @@ struct mtr_t {
 	{
 		ut_ad(!m_user_space_id);
 		ut_ad(!m_user_space);
-		ut_ad(!m_undo_space);
-		ut_ad(!m_sys_space);
 
 		ut_d(m_user_space_id = mtr.m_user_space_id);
 		m_user_space = mtr.m_user_space;
-		m_undo_space = mtr.m_undo_space;
-		m_sys_space = mtr.m_sys_space;
 	}
 
 	/** Set the tablespace associated with the mini-transaction
 	(needed for generating a MLOG_FILE_NAME record)
 	@param[in]	space_id	user or system tablespace ID
 	@return	the tablespace */
-	fil_space_t* set_named_space(ulint space_id)
+	fil_space_t* set_named_space_id(ulint space_id)
 	{
 		ut_ad(!m_user_space_id);
 		ut_d(m_user_space_id = space_id);
 		if (!space_id) {
-			return(set_sys_modified());
+			return fil_system.sys_space;
 		} else {
-			lookup_user_space(space_id);
+			ut_ad(m_user_space_id == space_id);
+			ut_ad(!m_user_space);
+			m_user_space = fil_space_get(space_id);
+			ut_ad(m_user_space);
 			return m_user_space;
 		}
 	}
@@ -253,7 +213,14 @@ struct mtr_t {
 	/** Set the tablespace associated with the mini-transaction
 	(needed for generating a MLOG_FILE_NAME record)
 	@param[in]	space	user or system tablespace */
-	void set_named_space(fil_space_t* space);
+	void set_named_space(fil_space_t* space)
+	{
+		ut_ad(!m_user_space_id);
+		ut_d(m_user_space_id = space->id);
+		if (space->id) {
+			m_user_space = space;
+		}
+	}
 
 #ifdef UNIV_DEBUG
 	/** Check the tablespace associated with the mini-transaction
@@ -261,38 +228,14 @@ struct mtr_t {
 	@param[in]	space	tablespace
 	@return whether the mini-transaction is associated with the space */
 	bool is_named_space(ulint space) const;
+	/** Check the tablespace associated with the mini-transaction
+	(needed for generating a MLOG_FILE_NAME record)
+	@param[in]	space	tablespace
+	@return whether the mini-transaction is associated with the space */
+	bool is_named_space(const fil_space_t* space) const;
 #endif /* UNIV_DEBUG */
 
-	/** Read 1 - 4 bytes from a file page buffered in the buffer pool.
-	@param ptr	pointer from where to read
-	@param type)	MLOG_1BYTE, MLOG_2BYTES, MLOG_4BYTES
-	@return	value read */
-	inline ulint read_ulint(const byte* ptr, mlog_id_t type) const
-		MY_ATTRIBUTE((warn_unused_result));
-
-	/** Locks a rw-latch in S mode.
-	NOTE: use mtr_s_lock().
-	@param lock	rw-lock
-	@param file	file name from where called
-	@param line	line number in file */
-	inline void s_lock(rw_lock_t* lock, const char* file, unsigned line);
-
-	/** Locks a rw-latch in X mode.
-	NOTE: use mtr_x_lock().
-	@param lock	rw-lock
-	@param file	file name from where called
-	@param line	line number in file */
-	inline void x_lock(rw_lock_t* lock, const char*	file, unsigned line);
-
-	/** Locks a rw-latch in X mode.
-	NOTE: use mtr_sx_lock().
-	@param lock	rw-lock
-	@param file	file name from where called
-	@param line	line number in file */
-	inline void sx_lock(rw_lock_t* lock, const char* file, unsigned line);
-
 	/** Acquire a tablespace X-latch.
-	NOTE: use mtr_x_lock_space().
 	@param[in]	space_id	tablespace ID
 	@param[in]	file		file name from where called
 	@param[in]	line		line number in file
@@ -302,11 +245,60 @@ struct mtr_t {
 		const char*	file,
 		unsigned	line);
 
-	/** Exclusively aqcuire a tablespace latch.
-	@param space  tablespace
-	@param file   source code file name of the caller
-	@param line   source code line number */
-	void x_lock_space(fil_space_t *space, const char *file, unsigned line);
+	/** Acquire a shared rw-latch.
+	@param[in]	lock	rw-latch
+	@param[in]	file	file name from where called
+	@param[in]	line	line number in file */
+	void s_lock(rw_lock_t* lock, const char* file, unsigned line)
+	{
+		rw_lock_s_lock_inline(lock, 0, file, line);
+		memo_push(lock, MTR_MEMO_S_LOCK);
+	}
+
+	/** Acquire an exclusive rw-latch.
+	@param[in]	lock	rw-latch
+	@param[in]	file	file name from where called
+	@param[in]	line	line number in file */
+	void x_lock(rw_lock_t* lock, const char* file, unsigned line)
+	{
+		rw_lock_x_lock_inline(lock, 0, file, line);
+		memo_push(lock, MTR_MEMO_X_LOCK);
+	}
+
+	/** Acquire an shared/exclusive rw-latch.
+	@param[in]	lock	rw-latch
+	@param[in]	file	file name from where called
+	@param[in]	line	line number in file */
+	void sx_lock(rw_lock_t* lock, const char* file, unsigned line)
+	{
+		rw_lock_sx_lock_inline(lock, 0, file, line);
+		memo_push(lock, MTR_MEMO_SX_LOCK);
+	}
+
+	/** Acquire a tablespace S-latch.
+	@param[in]	space	tablespace
+	@param[in]	file	file name from where called
+	@param[in]	line	line number in file */
+	void s_lock_space(fil_space_t* space, const char* file, unsigned line)
+	{
+		ut_ad(space->purpose == FIL_TYPE_TEMPORARY
+		      || space->purpose == FIL_TYPE_IMPORT
+		      || space->purpose == FIL_TYPE_TABLESPACE);
+		s_lock(&space->latch, file, line);
+	}
+
+	/** Acquire a tablespace X-latch.
+	@param[in]	space	tablespace
+	@param[in]	file	file name from where called
+	@param[in]	line	line number in file */
+	void x_lock_space(fil_space_t* space, const char* file, unsigned line)
+	{
+		ut_ad(space->purpose == FIL_TYPE_TEMPORARY
+		      || space->purpose == FIL_TYPE_IMPORT
+		      || space->purpose == FIL_TYPE_TABLESPACE);
+		memo_push(space, MTR_MEMO_SPACE_X_LOCK);
+		rw_lock_x_lock_inline(&space->latch, 0, file, line);
+	}
 
 	/** Release an object in the memo stack.
 	@param object	object
@@ -434,16 +426,11 @@ struct mtr_t {
 	static inline bool is_block_dirtied(const buf_block_t* block)
 		MY_ATTRIBUTE((warn_unused_result));
 
-	/** Check if we are holding a block latch in exclusive mode
-	@param block  buffer pool block to search for */
-	bool have_x_latch(const buf_block_t& block) const;
+	/** Determine the added buffer fix count of a block.
+	@param block block to be checked
+	@return number of buffer count added by this mtr */
+	uint32_t get_fix_count(const buf_block_t *block);
 private:
-	/** Look up the system tablespace. */
-	void lookup_sys_space();
-	/** Look up the user tablespace.
-	@param[in]	space_id	tablespace ID  */
-	void lookup_user_space(ulint space_id);
-
 	/** Prepare to write the mini-transaction log to the redo log buffer.
 	@return number of bytes to write in finish_write() */
 	inline ulint prepare_write();
@@ -485,10 +472,6 @@ private:
 #endif /* UNIV_DEBUG */
 	/** User tablespace that is being modified by the mini-transaction */
 	fil_space_t*	m_user_space;
-	/** Undo tablespace that is being modified by the mini-transaction */
-	fil_space_t*	m_undo_space;
-	/** System tablespace if being modified by the mini-transaction */
-	fil_space_t*	m_sys_space;
 
 	/** State of the transaction */
 	mtr_state_t	m_state;
