@@ -37,6 +37,8 @@ Created 12/15/2009	Jimmy Yang
 #endif /* __STDC_LIMIT_MACROS */
 
 #include <stdint.h>
+#include "my_atomic.h"
+#include "my_atomic_wrapper.h"
 
 /** Possible status values for "mon_status" in "struct monitor_value" */
 enum monitor_running_status {
@@ -177,7 +179,6 @@ enum monitor_id_t {
 	MONITOR_OVLD_INDEX_PAGES_WRITTEN,
 	MONITOR_OVLD_NON_INDEX_PAGES_WRITTEN,
 	MONITOR_OVLD_PAGES_READ,
-	MONITOR_OVLD_PAGES0_READ,
 	MONITOR_OVLD_INDEX_SEC_REC_CLUSTER_READS,
 	MONITOR_OVLD_INDEX_SEC_REC_CLUSTER_READS_AVOIDED,
 	MONITOR_OVLD_BYTE_READ,
@@ -300,7 +301,6 @@ enum monitor_id_t {
 	MONITOR_TRX_COMMIT_UNDO,
 	MONITOR_TRX_ROLLBACK,
 	MONITOR_TRX_ROLLBACK_SAVEPOINT,
-	MONITOR_TRX_ROLLBACK_ACTIVE,
 	MONITOR_TRX_ACTIVE,
 	MONITOR_RSEG_HISTORY_LEN,
 	MONITOR_NUM_UNDO_SLOT_USED,
@@ -656,7 +656,7 @@ Use MONITOR_DEC if appropriate mutex protection exists.
 
 #ifdef HAVE_valgrind
 # define MONITOR_CHECK_DEFINED(value) do {	\
-    mon_type_t m __attribute__((unused))= value;        \
+	mon_type_t m = value;			\
 	MEM_CHECK_DEFINED(&m, sizeof m);	\
 } while (0)
 #else /* HAVE_valgrind */
