@@ -28,9 +28,9 @@ Created 11/26/1995 Heikki Tuuri
 
 #include "buf0buf.h"
 #include "buf0flu.h"
+#include "fsp0sysspace.h"
 #include "page0types.h"
 #include "mtr0log.h"
-#include "row0trunc.h"
 #include "log0recv.h"
 
 /** Iterate over a memo block in reverse. */
@@ -167,7 +167,7 @@ struct FindPage
 			slot->object);
 
 		if (m_ptr < block->frame
-		    || m_ptr >= block->frame + block->page.size.logical()) {
+		    || m_ptr >= block->frame + srv_page_size) {
 			return(true);
 		}
 
@@ -234,7 +234,7 @@ static void memo_slot_release(mtr_memo_slot_t *slot)
   case MTR_MEMO_PAGE_X_FIX:
     buf_block_t *block= reinterpret_cast<buf_block_t*>(slot->object);
     buf_page_release_latch(block, slot->type);
-    buf_block_unfix(block);
+    block->unfix();
     break;
   }
   slot->object= NULL;
@@ -277,7 +277,7 @@ struct ReleaseLatches {
     case MTR_MEMO_PAGE_X_FIX:
       buf_block_t *block= reinterpret_cast<buf_block_t*>(slot->object);
       buf_page_release_latch(block, slot->type);
-      buf_block_unfix(block);
+      block->unfix();
       break;
     }
     slot->object= NULL;
@@ -400,7 +400,7 @@ mtr_write_log(
 	ut_ad(!recv_no_log_write);
 	DBUG_PRINT("ib_log",
 		   (ULINTPF " extra bytes written at " LSN_PF,
-		    len, log_sys->lsn));
+		    len, log_sys.lsn));
 
 	log_reserve_and_open(len);
 	log->for_each_block(write_log);
@@ -422,8 +422,6 @@ void mtr_t::start()
   m_log_mode= MTR_LOG_ALL;
   ut_d(m_user_space_id= TRX_SYS_SPACE);
   m_user_space= NULL;
-  m_undo_space= NULL;
-  m_sys_space= NULL;
   m_state= MTR_STATE_ACTIVE;
   m_flush_observer= NULL;
   m_commit_lsn= 0;
@@ -528,7 +526,7 @@ mtr_t::commit_checkpoint(
 	if (write_mlog_checkpoint) {
 		DBUG_PRINT("ib_log",
 			   ("MLOG_CHECKPOINT(" LSN_PF ") written at " LSN_PF,
-			    checkpoint_lsn, log_sys->lsn));
+			    checkpoint_lsn, log_sys.lsn));
 	}
 }
 
@@ -540,12 +538,7 @@ mtr_t::commit_checkpoint(
 bool
 mtr_t::is_named_space(ulint space) const
 {
-	ut_ad(!m_sys_space || m_sys_space->id == TRX_SYS_SPACE);
-	ut_ad(!m_undo_space || m_undo_space->id != TRX_SYS_SPACE);
 	ut_ad(!m_user_space || m_user_space->id != TRX_SYS_SPACE);
-	ut_ad(!m_sys_space || m_sys_space != m_user_space);
-	ut_ad(!m_sys_space || m_sys_space != m_undo_space);
-	ut_ad(!m_user_space || m_user_space != m_undo_space);
 
 	switch (get_log_mode()) {
 	case MTR_LOG_NONE:
@@ -560,6 +553,26 @@ mtr_t::is_named_space(ulint space) const
 	ut_error;
 	return(false);
 }
+/** Check if a tablespace is associated with the mini-transaction
+(needed for generating a MLOG_FILE_NAME record)
+@param[in]	space	tablespace
+@return whether the mini-transaction is associated with the space */
+bool mtr_t::is_named_space(const fil_space_t* space) const
+{
+  ut_ad(!m_user_space || m_user_space->id != TRX_SYS_SPACE);
+
+  switch (get_log_mode()) {
+  case MTR_LOG_NONE:
+  case MTR_LOG_NO_REDO:
+    return true;
+  case MTR_LOG_ALL:
+  case MTR_LOG_SHORT_INSERTS:
+    return m_user_space == space || is_predefined_tablespace(space->id);
+  }
+
+  ut_error;
+  return false;
+}
 #endif /* UNIV_DEBUG */
 
 /** Acquire a tablespace X-latch.
@@ -576,80 +589,22 @@ mtr_t::x_lock_space(ulint space_id, const char* file, unsigned line)
 	ut_ad(is_active());
 
 	if (space_id == TRX_SYS_SPACE) {
-		space = m_sys_space;
-
-		if (!space) {
-			space = m_sys_space = fil_space_get(space_id);
-		}
+		space = fil_system.sys_space;
 	} else if ((space = m_user_space) && space_id == space->id) {
-	} else if ((space = m_undo_space) && space_id == space->id) {
-	} else if (get_log_mode() == MTR_LOG_NO_REDO) {
-		space = fil_space_get(space_id);
-		ut_ad(space->purpose == FIL_TYPE_TEMPORARY
-		      || space->purpose == FIL_TYPE_IMPORT
-		      || space->redo_skipped_count > 0
-		      || srv_is_tablespace_truncated(space->id));
 	} else {
-		/* called from trx_rseg_create() */
-		space = m_undo_space = fil_space_get(space_id);
+		space = fil_space_get(space_id);
+		ut_ad(get_log_mode() != MTR_LOG_NO_REDO
+		      || space->purpose == FIL_TYPE_TEMPORARY
+		      || space->purpose == FIL_TYPE_IMPORT
+		      || space->redo_skipped_count > 0);
 	}
 
 	ut_ad(space);
 	ut_ad(space->id == space_id);
 	x_lock_space(space, file, line);
-	ut_ad(space->purpose == FIL_TYPE_TEMPORARY
-	      || space->purpose == FIL_TYPE_IMPORT
-	      || space->purpose == FIL_TYPE_TABLESPACE);
 	return(space);
 }
 
-/** Exclusively aqcuire a tablespace latch.
-@param space  tablespace
-@param file   source code file name of the caller
-@param line   source code line number */
-void mtr_t::x_lock_space(fil_space_t *space, const char *file, unsigned line)
-{
-  rw_lock_x_lock_inline(&space->latch, 0, file, line);
-  memo_push(space, MTR_MEMO_SPACE_X_LOCK);
-}
-
-/** Look up the system tablespace. */
-void
-mtr_t::lookup_sys_space()
-{
-	ut_ad(!m_sys_space);
-	m_sys_space = fil_space_get(TRX_SYS_SPACE);
-	ut_ad(m_sys_space);
-}
-
-/** Look up the user tablespace.
-@param[in]	space_id	tablespace ID */
-void
-mtr_t::lookup_user_space(ulint space_id)
-{
-	ut_ad(space_id != TRX_SYS_SPACE);
-	ut_ad(m_user_space_id == space_id);
-	ut_ad(!m_user_space);
-	m_user_space = fil_space_get(space_id);
-	ut_ad(m_user_space);
-}
-
-/** Set the tablespace associated with the mini-transaction
-(needed for generating a MLOG_FILE_NAME record)
-@param[in]	space	user or system tablespace */
-void
-mtr_t::set_named_space(fil_space_t* space)
-{
-	ut_ad(m_user_space_id == TRX_SYS_SPACE);
-	ut_d(m_user_space_id = space->id);
-	if (space->id == TRX_SYS_SPACE) {
-		ut_ad(!m_sys_space || m_sys_space == space);
-		m_sys_space = space;
-	} else {
-		m_user_space = space;
-	}
-}
-
 /** Release an object in the memo stack.
 @return true if released */
 bool
@@ -704,7 +659,7 @@ inline ulint mtr_t::prepare_write()
 		ut_ad(m_log_mode == MTR_LOG_NO_REDO);
 		ut_ad(m_log.size() == 0);
 		log_mutex_enter();
-		m_commit_lsn = log_sys->lsn;
+		m_commit_lsn = log_sys.lsn;
 		return 0;
 	}
 
@@ -713,8 +668,8 @@ inline ulint mtr_t::prepare_write()
 	ut_ad(len > 0);
 	ut_ad(n_recs > 0);
 
-	if (len > log_sys->buf_size / 2) {
-		log_buffer_extend((len + 1) * 2);
+	if (len > srv_log_buffer_size / 2) {
+		log_buffer_extend(ulong((len + 1) * 2));
 	}
 
 	ut_ad(m_n_log_recs == n_recs);
@@ -798,31 +753,7 @@ inline lsn_t mtr_t::finish_write(ulint len)
 	return start_lsn;
 }
 
-/** Release the free extents that was reserved using
-fsp_reserve_free_extents().  This is equivalent to calling
-fil_space_release_free_extents().  This is intended for use
-with index pages.
-@param[in]	n_reserved	number of reserved extents */
-void
-mtr_t::release_free_extents(ulint n_reserved)
-{
-  fil_space_t *space= m_user_space;
-
-  ut_ad(!m_undo_space);
-
-  if (space)
-    ut_ad(m_user_space->id == m_user_space_id);
-  else
-  {
-    ut_ad(m_sys_space->id == TRX_SYS_SPACE);
-    space= m_sys_space;
-  }
-
-  ut_ad(memo_contains(get_memo(), space, MTR_MEMO_SPACE_X_LOCK));
-  space->release_free_extents(n_reserved);
-}
-
-int32_t mtr_t::get_fix_count(buf_block_t *block)
+uint32_t mtr_t::get_fix_count(const buf_block_t *block)
 {
   Iterate<FindBlock> iteration((FindBlock(block)));
   if (m_memo.for_each_block(iteration))
